import java.io.InputStream;
import java.io.IOException;

final class Tokenize {
    private InputStream in;

    private int readAhead;

    public Tokenize(InputStream in) throws IOException
    {
	this.in = in;
	readAhead = in.read();
    }

    public int nextTokenType() throws IOException
    {
	if(readAhead == '<') {
	    int tag;
	    int ch;
	    switch(ch = in.read()) {
	    case 'B': tag = Tag.b; break;
	    case 'E': in.read(); /* M */ tag = Tag.em; break;
	    case 'I': tag = Tag.i; break;
	    case 'P': in.read(); /* L */ tag = Tag.pl; break;
	    case 'S': tag = Tag.s; break;
	    case 'T': in.read(); /* T */  tag = Tag.tt; break;
	    case 'U': tag = Tag.u1; break;
	    case '0': tag = Tag.s0; break;
	    case '1': tag = Tag.s1; break;
	    case '2': tag = Tag.s2; break;
	    case '3': tag = Tag.s3; break;
	    case '4': tag = Tag.s4; break;
	    case '5': tag = Tag.s5; break;
	    case '6': tag = Tag.s6; break;
	    case '7': tag = Tag.s7; break;
	    case '8': tag = Tag.s8; break;
	    case '9': tag = Tag.s9; break;
	    case 'r': tag = Tag.rc; break;
	    case 'g': tag = Tag.gc; break;
	    case 'b': tag = Tag.bc; break;
	    case 'c': tag = Tag.cc; break;
	    case 'm': tag = Tag.mc; break;
	    case 'y': tag = Tag.yc; break;
	    case 'k': tag = Tag.kc; break;
	    case 'w': tag = Tag.wc; break;
	    case '/':
		tag = Tag.lexerExit;
		switch(in.read()) {
		case 'E': in.read(); break;
		case 'P': in.read(); break;
		case 'T': in.read(); break;
		}
		break;
	    default:
		System.err.println("lexer error: unrecognized tag " + ch);
		return Tag.lexerEOF;
	    }

	    if(in.read() != '>') {
		System.err.println("lexer error: expecting '>'");
		return Tag.lexerEOF;
	    }
	    readAhead = in.read();
	    return tag;
	} else if(readAhead == -1) {
	    return Tag.lexerEOF;
	} else {
	    /* leave readAhead in place */
	    return Tag.string;
	}
    }

    public StringBuffer nextTokenString() throws IOException
    {
	StringBuffer sb = new StringBuffer();
	int next = readAhead;
	do {
	    sb.append((char) next);
	    next = in.read();
	} while(next != -1 && next != '<');

	readAhead = next;
	return sb;
    }
}

public final class Parser {
    private Tokenize tok;

    public Parser(InputStream in) throws IOException
    {
	this.tok = new Tokenize(in);
    }

    private int stringCont = 0 ;
    private boolean wasSpace = false ;
    private Node prevStringNode = null ;

    private static StringBuffer emptyString = new StringBuffer() ;
    private Node makeFilteredStringNode (int c, StringBuffer s){
	// filter out duplicate whitespace
	StringBuffer s2 ;
	boolean createNode ;
	boolean pullRight = false ;
	boolean allSpaces = true ;
	int i = 0 ;
	char ch ;
	StringBuffer sb = null ;

	sb = prevStringNode.getStringBuffer();

	if(c == stringCont){
	    s2 = sb ;
	    createNode = false ;
	}else if(Tag.spaceContext(c) != Tag.spaceContext(stringCont)){
	    wasSpace = false ;
	    s2 = emptyString ;
	    createNode = true ;
	}else{
	    s2 = emptyString ;
	    
	    for(i = 0; i < s.length(); i++){
		ch = s.charAt(i);
		if(! Character.isWhitespace(ch)){
		    break ;
		}
		if(Tag.isTt(c)) {
		    sb.append(ch);
		} else if(! wasSpace) {
		    sb.append(' ');
		}

		wasSpace = true ;

	    }

	    pullRight = Tag.isSpaceString(prevStringNode.getTag()) && (i < s.length()) ;
	    createNode = i < s.length() ;
	}

	for( ; i < s.length(); i++){
	    ch = s.charAt(i);

	    if(Character.isWhitespace(ch)){
		if(Tag.isTt(c)) {
		    s2.append(ch);
		} else if(! wasSpace){
		    s2.append(' ');
		}
		wasSpace = true ;
	    }else{
		allSpaces = false ;
		s2.append(ch);
		wasSpace = false ;
	    }

	}

	if(pullRight){
	    sb.append(s2) ;
	    s2 = sb ;
	    prevStringNode.clear();
	}

	// only add a node if string isn't empty
	if(s2.length() != 0 && createNode){
	    emptyString = new StringBuffer(); // refresh emptyString
	    Node n = new Node(s2, allSpaces);
	    StringBuffer foooo = prevStringNode.getStringBuffer();
	    if(foooo != null) {
		prevStringNode.setString(foooo.toString());
	    }
	    prevStringNode = n ;
	    stringCont = c ;
	    return n ;
	}else{
	    return null ;
	}

    }

    private int depth ;

    private Node recParse
	(
	 int c                  // current context
	 ) throws IOException
    {
	int tag = tok.nextTokenType() ;
	if(tag == Tag.lexerEOF){
	    depth-- ;
	    return null ;

	}else if(tag == Tag.string) {
	    return makeFilteredStringNode(c, tok.nextTokenString());

	} else if(tag == Tag.lexerExit) {
	    /* close tag */
	    depth-- ;
	    return null ;

	} else {
	    /* open tag */
	    int d = depth ;
	    depth++ ;

	    if(Tag.isU1(tag)){
		if(Tag.isU3(c)){
		    // already at U3, so this U is redundant
		    return null ;
		}else if(Tag.isU2(c)){
		    tag = (tag & ~Tag.u1) | Tag.u3 ;
		}else if(Tag.isU1(c)){
		    tag = (tag & ~Tag.u1) | Tag.u2 ;
		}
	    }

	    // can't be done until _after_ the correct u level is given to the tag
	    int c2 = Tag.composeOntoContext(c, tag);

	    if(c2 == c){
		return null ;
	    }else{
		Node curr = null ;
		Node n ;

		while(true){
		    n = recParse(c2);
		    if(n != null){
			if(curr == null){
			    curr = new Node(tag);
			}
			curr.add(n);
		    }else if(d == depth){
			break ;
		    }
		}

		return curr ;
	    }

	}
	
    }


    public Node superParse() throws IOException
    {
	Node rootNode = new Node(Tag.empty);
	Node n ;

	stringCont = Tag.topLevel ;
	wasSpace = false ;
	Node dummyLeadStringNode = new Node(new StringBuffer(), false);
	prevStringNode = dummyLeadStringNode ;

	rootNode.add(dummyLeadStringNode);

	depth = 0 ;

	while(true){
	    n = recParse(Tag.topLevel);
	    if(n != null){
		rootNode.add(n);
	    }else{
		if(depth == -1){
		    break ;
		}
	    }
	}

	prevStringNode.setString(prevStringNode.getStringBuffer().toString().intern());

	if(dummyLeadStringNode.getString().length() == 0){
	    // didn't use the dummy lead string.  kill it.
	    dummyLeadStringNode.clear();
	}

	return rootNode ;

    }
    /*
    public Node parse() throws IOException
    { 
	Node n = new Node(Tag.empty);

	int tag;
	int underlines=0;
	while((tag = tok.nextTokenType()) != Tag.lexerEOF) {
	    if(tag == Tag.string) {
		n.add(new Node(tok.nextTokenString(), false));
	    } else if(tag == Tag.lexerExit) {
		if(underlines == 1 && Tag.isU1(n.getTag())) {
		    underlines--;
		} else if(underlines == 2 && Tag.isU2(n.getTag())) {
		    underlines--;
		} else if(underlines == 3 && Tag.isU3(n.getTag())) {
		    underlines--;
		}
		n = n.getParent();
	    } else { // "normal" tag
		if(tag == Tag.u1) {
		    if(underlines < 3) {
			underlines++;
			tag = (underlines == 1 ? Tag.u1 :
			       (underlines == 2 ? Tag.u2 : Tag.u3));
		    } else {
			tag = Tag.empty;
		    }
		}
		Node nn = new Node(tag);
		n.add(nn);
		n = nn;
	    }
	}

	if(n.getParent() != null) {
	    System.err.println("parser error: not back at root on EOF");
	}

	return n;
    }
*/
}
